# DISABLE CUDA

import os
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

# IMPORT TENSORFLOW AND CONFIGURE SETTINGS

import tensorflow as tf

# ENABLE KEYBOARD INTERRUPTIONS TO CONTROL RENDERING WITH KEY 'r'

"""
from pynput import keyboard

def on_press(key):
    try:
        if key.char == 'r':
            print("Toogling Rendering... ")
            train_py_env.toogle_rendering()
            eval_py_env.toogle_rendering()
    except AttributeError:
        print('pynput error!!!')

listener = keyboard.Listener(on_press=on_press)

# listener.start() # Not implemented yet (it was, but crashed a lot, don't know why)
"""

# OFF_POLICY TIME_STEP BASED EXPERIMENT GENERIC HYPERPARAMETERS

import datetime

root_dir = os.path.join("logs", datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))

num_eval_episodes = 3
summary_interval = 100
env_name = "BipedalWalker-v3"
env_action_repeat_times = 4
debug_summaries=False
summarize_grads_and_vars=False
replay_buffer_capacity=5000
initial_collect_steps=1000
collect_steps_per_iteration=1
use_tf_functions=True
sample_batch_size = 256
num_iterations=50000
train_steps_per_iteration=1
log_interval=1000
eval_interval=1000
train_checkpoint_interval=5000
policy_checkpoint_interval=5000
replay_buffer_checkpoint_interval=5000

# SAC AND NAF ANALOG HYPERPARAMETERS
learning_rate=3e-4
target_update_tau=0.005
target_update_period=1
td_errors_loss_fn=tf.math.squared_difference
gamma=0.99


# NAF SPECIFIC HYPERPARAMETERS

from tf_agents.networks import encoding_network

naf_root_dir = os.path.join(root_dir, "naf")
preprocessing_conv_layer_params = None
preprocessing_conv_type = encoding_network.CONV_TYPE_1D
preprocessing_fc_layer_params = (256, )
preprocessing_dropout_layer_params = None
v_network_fc_layer_params = (256, )
v_network_dropout_layer_params = None
l_network_fc_layer_params = (256, )
l_network_dropout_layer_params = None
policy_network_fc_layer_params = (256, )
policy_network_dropout_layer_params = None
policy_network_uses_shared_preprocessing_network = True
noise_factor=0.1

# SAC SPECIFIC HYPERPARAMETERS

sac_root_dir = os.path.join(root_dir, "sac")
actor_fc_layers=(256, 256)
critic_obs_fc_layers=None
critic_action_fc_layers=None
critic_joint_fc_layers=(256, 256)
actor_learning_rate=learning_rate
critic_learning_rate=learning_rate
alpha_learning_rate=learning_rate
reward_scale_factor=0.1
gradient_clipping=None


# RANDOM POLICY AGENT (RPA) SPECIFIC HYPERPARAMETERS

rpa_root_dir = os.path.join(root_dir, "rpa")


# SETUP BASE EXPERIMENTS

from experiments.naf_experiment import NafExperiment

naf_base_experiment = NafExperiment(
    root_dir = naf_root_dir,
    num_eval_episodes = num_eval_episodes,
    summary_interval = summary_interval,
    env_name = env_name,
    env_action_repeat_times = env_action_repeat_times,
    preprocessing_conv_layer_params = preprocessing_conv_layer_params,
    preprocessing_conv_type = preprocessing_conv_type,
    preprocessing_fc_layer_params = preprocessing_fc_layer_params,
    preprocessing_dropout_layer_params = preprocessing_dropout_layer_params,
    v_network_fc_layer_params = v_network_fc_layer_params,
    v_network_dropout_layer_params = v_network_dropout_layer_params,
    l_network_fc_layer_params = l_network_fc_layer_params,
    l_network_dropout_layer_params = l_network_dropout_layer_params,
    policy_network_fc_layer_params = policy_network_fc_layer_params,
    policy_network_dropout_layer_params = policy_network_dropout_layer_params,
    policy_network_uses_shared_preprocessing_network = policy_network_uses_shared_preprocessing_network,
    learning_rate=learning_rate,
    target_update_tau=target_update_tau,
    target_update_period=target_update_period,
    td_errors_loss_fn=td_errors_loss_fn,
    gamma=gamma,
    noise_factor=noise_factor,
    debug_summaries=debug_summaries,
    summarize_grads_and_vars=summarize_grads_and_vars,
    replay_buffer_capacity=replay_buffer_capacity,
    initial_collect_steps=initial_collect_steps,
    collect_steps_per_iteration=collect_steps_per_iteration,
    use_tf_functions=use_tf_functions,
    sample_batch_size = sample_batch_size,
    num_iterations=num_iterations,
    train_steps_per_iteration=train_steps_per_iteration,
    log_interval=log_interval,
    eval_interval=eval_interval,
    train_checkpoint_interval=train_checkpoint_interval,
    policy_checkpoint_interval=policy_checkpoint_interval,
    replay_buffer_checkpoint_interval=replay_buffer_checkpoint_interval,
    name="naf_base_experiment"
    )

from experiments.sac_experiment import SacExperiment

sac_base_experiment = SacExperiment(
    root_dir = sac_root_dir,
    num_eval_episodes = num_eval_episodes,
    summary_interval = summary_interval,
    env_name = env_name,
    env_action_repeat_times = env_action_repeat_times,
    actor_fc_layers=actor_fc_layers,
    critic_obs_fc_layers=critic_obs_fc_layers,
    critic_action_fc_layers=critic_action_fc_layers,
    critic_joint_fc_layers=critic_joint_fc_layers,
    actor_learning_rate=actor_learning_rate,
    critic_learning_rate=critic_learning_rate,
    alpha_learning_rate=alpha_learning_rate,
    target_update_tau=target_update_tau,
    target_update_period=target_update_period,
    td_errors_loss_fn=td_errors_loss_fn,
    gamma=gamma,
    reward_scale_factor=reward_scale_factor,
    gradient_clipping=gradient_clipping,
    debug_summaries=debug_summaries,
    summarize_grads_and_vars=summarize_grads_and_vars,
    replay_buffer_capacity=replay_buffer_capacity,
    initial_collect_steps=initial_collect_steps,
    collect_steps_per_iteration=collect_steps_per_iteration,
    use_tf_functions=use_tf_functions,
    sample_batch_size = sample_batch_size,
    num_iterations=num_iterations,
    train_steps_per_iteration=train_steps_per_iteration,
    log_interval=log_interval,
    eval_interval=eval_interval,
    train_checkpoint_interval=train_checkpoint_interval,
    policy_checkpoint_interval=policy_checkpoint_interval,
    replay_buffer_checkpoint_interval=replay_buffer_checkpoint_interval,
    name="sac_base_experiment"
    )

from experiments.rnd_experiment import RndExperiment

rnd_base_experiment = RndExperiment(
    root_dir = rpa_root_dir,
    num_eval_episodes = num_eval_episodes,
    summary_interval = summary_interval,
    env_name = env_name,
    env_action_repeat_times = env_action_repeat_times,
    debug_summaries=debug_summaries,
    summarize_grads_and_vars=summarize_grads_and_vars,
    replay_buffer_capacity=replay_buffer_capacity,
    initial_collect_steps=initial_collect_steps,
    collect_steps_per_iteration=collect_steps_per_iteration,
    use_tf_functions=use_tf_functions,
    sample_batch_size = sample_batch_size,
    num_iterations=num_iterations,
    train_steps_per_iteration=train_steps_per_iteration,
    log_interval=log_interval,
    eval_interval=eval_interval,
    train_checkpoint_interval=train_checkpoint_interval,
    policy_checkpoint_interval=policy_checkpoint_interval,
    replay_buffer_checkpoint_interval=replay_buffer_checkpoint_interval,
    name="random_policy_agent_base_experiment"
    )

# GENERATE THE WANTED EXPERIMENTS

experiment_list = []

for times in (4, 8):
    experiment = naf_base_experiment.copy()
    experiment.name = "naf_env_hardcore_times_{}".format(times)
    experiment.env_name = "BipedalWalkerHardcore-v3"
    experiment.env_action_repeat_times = times
    experiment_list.append(experiment)

# PREPARE FOR LAUNCHING

def launch_experiment(experiment):
    experiment.launch()

# LAUNCH EXPERIMENTS

import multiprocessing

if(len(experiment_list) == 1):
    # Single process execution
    experiment_list[0].launch()
else:
    # Multiple processes execution (ideally, no more than the number of cores of your CPU)

    # TODO launch only two experiments simultaneously, when one finish, launch the next one
    
    processes_list = list()

    if __name__ == "__main__":

        for experiment in experiment_list:
            process = multiprocessing.Process(
                target=launch_experiment, args=[experiment])
            processes_list.append(process)
            process.start()
            print("Process \"" + str(experiment.name) + "\" started with pid = " + str(process.pid))
